/***************************************************************************
 *   Copyright (C) 2012 by santiago González                               *
 *   santigoro@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, see <http://www.gnu.org/licenses/>.  *
 *                                                                         *
 ***************************************************************************/

#include "e-node.h"
#include "simulator.h"
#include "e-element.h"

eNode::eNode( QString id )
{
    m_id = id;
    m_nodeNum = 0;
    m_volt    = 0;
    m_isBus = false;

    initialize();

    Simulator::self()->addToEnodeList( this );
}
eNode::~eNode(){}

void eNode::initialize()
{
    m_switched     = false;
    m_single       = false;
    m_changed      = false;
    m_currChanged  = false;
    m_admitChanged = false;

    m_changedFast.clear();
    m_nonLinear.clear();
    m_reactiveList.clear();

    m_admitVector.clear();
    m_currentVector.clear();
    m_nodeVector.clear();

    m_eBusPinList.clear();

    m_totalAdmit = 0;
    m_volt = 0;
}

int eNode::addLink( int nodeNum ) // Add node at other side of pin
{
    int index = m_admitVector.size();

    m_nodeVector.emplace_back( nodeNum );
    m_admitVector.emplace_back( 0 );
    m_currentVector.emplace_back( 0 );

    return index;
}

void eNode::stampCurrent( int nodeNum, double data )
{
    if( (uint)nodeNum >= m_currentVector.size() ) return;
    //if( m_nodeVector[nodeNum] == m_nodeNum  ) return; // Be sure msg doesn't come from this node

    m_currentVector[nodeNum] = data;
    m_currChanged = true;

    if( !m_changed )
    {
        m_changed = true;
        Simulator::self()->addToChangedNodeList( this );
    }
}

void eNode::stampAdmitance( int nodeNum, double data )
{
    if( (uint)nodeNum >= m_admitVector.size() ) return;
    //if( m_nodeVector[nodeNum] == m_nodeNum  ) return; // Be sure msg doesn't come from this node

    //double adm = m_admitVector[nodeNum];
    //m_admit[nodeNum] += data-adm;
    m_totalAdmit += data-m_admitVector[nodeNum];

    m_admitVector[nodeNum] = data;
    m_admitChanged = true;

    if( !m_changed )
    {
        m_changed = true;
        Simulator::self()->addToChangedNodeList( this );
    }
}

void eNode::stampMatrix()
{
    if( m_nodeNum == 0 ) return;

    m_changed = false;

    if( m_admitChanged )
    {
        if( !m_single || m_switched )
        {
            m_admit.clear();

            for( uint i=0; i<m_admitVector.size(); ++i )
            {
                int enode  = m_nodeVector[i];
                m_admit[enode] += m_admitVector[i];
            }
            stampAdmit();
        }
        m_admitChanged = false;
    }

    if( m_currChanged )
    {
        m_totalCurr = 0;
        for( double current : m_currentVector ) m_totalCurr += current;

        if( !m_single || m_switched ) CircMatrix::self()->stampCoef( m_nodeNum, m_totalCurr );

        m_currChanged  = false;
    }
    if( m_single ) solveSingle();
}

void eNode::stampAdmit()
{
    int nonCero = 0;
    QHashIterator<int, double> ai(m_admit); // iterate admitance hash: eNode-Admit
    while ( ai.hasNext() )
    {
        ai.next();
        int    enode = ai.key();
        double admit = ai.value();

        if( enode>0 ) CircMatrix::self()->stampMatrix( m_nodeNum, enode, -admit );

        if( m_switched )                       // Find open/close events
        {
            if( admit > 0 ) nonCero++;
            double admitP = m_admitPrev[enode];

            if(( admit != admitP )
              &&((admit==0)||(admitP==0))) CircMatrix::self()->setCircChanged();
        }
    }
    if( m_switched )
    {
        m_admitPrev = m_admit;
        if( nonCero < 2 ) m_totalAdmit += 1e-12; //pnpBias example error
    }
    CircMatrix::self()->stampMatrix( m_nodeNum, m_nodeNum, m_totalAdmit );
}

void eNode::solveSingle()
{
    double volt = 0;

    if( m_totalAdmit > 0 ) volt = m_totalCurr/m_totalAdmit;
    setVolt( volt );
}

QList<int> eNode::getConnections()
{
    QList<int> cons;
    for( int nodeNum : m_nodeVector )
    {
        if( m_admit[nodeNum] > 0
        && !cons.contains(nodeNum)) cons.append( nodeNum );
    }
    return cons;
}

void  eNode::setVolt( double v )
{
    if( m_volt != v ) //( fabs(m_volt-v) > 1e-9 ) //
    {
        m_voltChanged = true;
        m_volt = v;

        for( eElement* el : m_changedFast )  Simulator::self()->addToChangedFast( el ); // el->setVChanged();
        for( eElement* el : m_reactiveList ) Simulator::self()->addToReactiveList( el );
        for( eElement* el : m_nonLinear )    Simulator::self()->addToNoLinList( el );
    }
}

void eNode::setIsBus( bool bus )
{
    m_isBus = bus;

    Simulator::self()->remFromEnodeList( this, /*delete=*/ false );
    Simulator::self()->addToEnodeBusList( this );
}

void eNode::createBus()
{
    int busSize = m_eBusPinList.size();

    for( int i=0; i<busSize; i++ )
    {
        QList<ePin*> pinList = m_eBusPinList.at( i );

        if( !pinList.isEmpty() )
        {
            eNode* enode = new eNode( m_id+"-eNode-"+QString::number( i ) );
            for( ePin* epin : pinList ) epin->setEnode( enode );
        }
    }
}

void eNode::addBusPinList( QList<ePin*> list, int line )
{
    int size = line+1;
    int busSize = m_eBusPinList.size();

    if( size > busSize )
    {
        for( int i=0; i<size-busSize; i++ )
        {
            QList<ePin*> newList;
            m_eBusPinList.append( newList );
        }
    }

    QList<ePin*> pinList = m_eBusPinList.at( line );
    for( ePin* epin : list )
        if( !pinList.contains( epin )) pinList.append( epin );

    m_eBusPinList.replace( line, pinList );
}

QList<ePin*> eNode::getEpins() { return m_ePinList; }

void eNode::addEpin( ePin* epin )
{ if( !m_ePinList.contains(epin)) m_ePinList.append(epin); }

void eNode::remEpin( ePin* epin )
{
    if( m_ePinList.contains(epin) ) m_ePinList.removeOne( epin );

    if( m_ePinList.isEmpty() ) // If No epins then remove this enode
    {
        if( m_isBus ) Simulator::self()->remFromEnodeBusList( this, true );
        else          Simulator::self()->remFromEnodeList( this, true );
    }
}

void eNode::setNodeNumber( int n ) { m_nodeNum = n; }

void eNode::addToChangedFast( eElement* el )
{ if( !m_changedFast.contains(el) ) m_changedFast.append(el); }

void eNode::remFromChangedFast( eElement* el )
{ m_changedFast.removeOne(el); }

void eNode::addToReactiveList( eElement* el )
{ if( !m_reactiveList.contains(el) ) m_reactiveList.append(el); }

void eNode::remFromReactiveList( eElement* el )
{ m_reactiveList.removeOne(el); }

void eNode::addToNoLinList( eElement* el )
{ if( !m_nonLinear.contains(el) ) m_nonLinear.append(el); }

void eNode::remFromNoLinList( eElement* el )
{ m_nonLinear.removeOne(el); }
int eNode::getNodeNumber() { return m_nodeNum; }

QString eNode::itemId() { return m_id; }
